from enum import Enum
from typing import Optional, List, Any

from octohook.models import (
    Repository,
    Organization,
    User,
    Comment,
    CheckRun,
    CheckSuite,
    Installation,
    DeployKey,
    Deployment,
    DeploymentStatus,
    Page,
    ShortRepository,
    Issue,
    Label,
    MarketplacePurchase,
    Team,
    Hook,
    Milestone,
    Membership,
    Package,
    PageBuild,
    ProjectCard,
    ProjectColumn,
    Project,
    PullRequest,
    Review,
    Release,
    VulnerabilityAlert,
    SecurityAdvisory,
    Sponsorship,
    Branch,
    StatusCommit,
    RawDict,
    ContentReference,
    Commit,
    CommitUser,
    _optional,
    ShortInstallation,
    Enterprise,
    Thread,
    Rule,
)


class BaseWebhookEvent:
    payload: dict
    action: Optional[str] = None
    sender: Optional[User]
    repository: Optional[Repository] = None
    organization: Optional[Organization] = None
    enterprise: Optional[Enterprise] = None

    def __init__(self, payload: dict):
        self.payload = payload
        self.action = payload.get("action")
        self.sender = _optional(payload, "sender", User)

        # Not present in GitHubAppAuthorizationEvent, InstallationEvent, SponsorshipEvent
        try:
            self.repository = Repository(payload.get("repository"))
        except AttributeError:
            pass

        # Only present in some events
        try:
            self.organization = Organization(payload.get("organization"))
        except AttributeError:
            pass

        self.enterprise = _optional(payload, "enterprise", Enterprise)


class BranchProtectionRuleEvent(BaseWebhookEvent):
    payload: dict
    rule: Rule
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.rule = Rule(payload.get("rule"))
        self.changes = _optional(payload, "changes", RawDict)


class CheckRunEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#checkrunevent
    """

    check_run: CheckRun

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.check_run = CheckRun(payload.get("check_run"))


class CheckSuiteEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#checksuiteevent
    """

    check_suite: CheckSuite

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.check_suite = CheckSuite(payload.get("check_suite"))


class CommitCommentEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#commitcommentevent
    """

    comment: Comment

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.comment = Comment(payload.get("comment"))


class ContentReferenceEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#contentreferenceevent
    """

    content_reference: ContentReference
    installation: ShortInstallation

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.content_reference = ContentReference(payload.get("content_reference"))
        self.installation = ShortInstallation(payload.get("installation"))


class CreateEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#createevent
    """

    ref: str
    ref_type: str
    master_branch: str
    description: str
    pusher_type: str

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.ref = payload.get("ref")
        self.ref_type = payload.get("ref_type")
        self.master_branch = payload.get("master_branch")
        self.description = payload.get("description")
        self.pusher_type = payload.get("pusher_type")


class DeleteEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#deleteevent
    """

    ref: str
    ref_type: str
    pusher_type: str

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.ref = payload.get("ref")
        self.ref_type = payload.get("ref_type")
        self.pusher_type = payload.get("pusher_type")


class DeployKeyEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#deploykeyevent
    """

    key: DeployKey

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.key = DeployKey(payload.get("key"))


class DeploymentEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#deploymentevent
    """

    deployment: Deployment

    def __init__(self, payload: dict):
        super().__init__(payload)

        self.deployment = Deployment(payload.get("deployment"))


class DeploymentStatusEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#deploymentstatusevent
    """

    deployment_status: DeploymentStatus
    deployment: Deployment

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.deployment_status = DeploymentStatus(payload.get("deployment_status"))
        self.deployment = Deployment(payload.get("deployment"))


class ForkEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#forkevent
    """

    forkee: Repository

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.forkee = Repository(payload.get("forkee"))


class GitHubAppAuthorizationEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#forkapplyevent
    """

    def __init__(self, payload: dict):
        super().__init__(payload)


class GollumEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#gollumevent
    """

    pages: List[Page]

    def __init__(self, payload: dict):
        super().__init__(payload)

        self.pages = [Page(page) for page in payload.get("pages")]


class InstallationEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#installationevent
    """

    installation: Installation
    repositories: List[ShortRepository]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.installation = Installation(payload.get("installation"))

        self.repositories = [
            ShortRepository(repo) for repo in payload.get("repositories", [])
        ]


class InstallationRepositoriesEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#installationrepositoriesevent
    """

    installation: Installation
    repository_selection: str
    repositories_added: List[ShortRepository]
    repositories_removed: List[ShortRepository]

    def __init__(self, payload: dict):
        super().__init__(payload)

        self.installation = Installation(payload.get("installation"))
        self.repository_selection = payload.get("repository_selection")
        self.repositories_added = [
            ShortRepository(repo) for repo in payload.get("repositories_added")
        ]
        self.repositories_removed = [
            ShortRepository(repo) for repo in payload.get("repositories_removed")
        ]


class IssueCommentEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#issuecommentevent
    """

    issue: Issue
    comment: Comment
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.issue = Issue(payload.get("issue"))
        self.comment = Comment(payload.get("comment"))
        self.changes = _optional(payload, "changes", RawDict)


class IssuesEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#issuesevent
    """

    issue: Issue
    changes: Optional[RawDict]
    label: Optional[Label]
    assignee: Optional[User]
    milestone: Optional[Milestone]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.issue = Issue(payload.get("issue"))
        self.changes = _optional(payload, "changes", RawDict)
        self.label = _optional(payload, "label", Label)
        self.assignee = _optional(payload, "assignee", User)
        self.milestone = _optional(payload, "milestone", Milestone)


class LabelEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#labelevent
    """

    label: Label
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.label = Label(payload.get("label"))
        self.changes = _optional(payload, "changes", RawDict)


class MarketplacePurchaseEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#marketplacepurchaseevent
    """

    effective_date: str
    marketplace_purchase: MarketplacePurchase

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.effective_date = payload.get("effective_date")
        self.marketplace_purchase = MarketplacePurchase(
            payload.get("marketplace_purchase")
        )


class MemberEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#memberevent
    """

    member: User

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.member = User(payload.get("member"))


class MembershipEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#membershipevent
    """

    scope: str
    member: User
    team: Team

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.scope = payload.get("scope")
        self.member = User(payload.get("member"))
        self.team = Team(payload.get("team"))


class MetaEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#metaevent
    """

    hook_id: int
    hook: Hook

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.hook_id = payload.get("hook_id")
        self.hook = Hook(payload.get("hook"))


class MilestoneEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#milestoneevent
    """

    milestone: Milestone
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.milestone = Milestone(payload.get("milestone"))
        self.changes = _optional(payload, "changes", RawDict)


class OrganizationEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#organizationevent
    """

    invitation: Optional[Any]
    membership: Membership

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.invitation = payload.get("invitation")  # TODO: What does this look like?
        self.membership = Membership(payload.get("membership"))


class OrgBlockEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#orgblockevent
    """

    blocked_user: User

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.blocked_user = User(payload.get("blocked_user"))


class PackageEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#packageevent
    """

    package: Package

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.package = Package(payload.get("package"))


class PageBuildEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#pagebuildevent
    """

    id: int
    build: PageBuild

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.id = payload.get("id")
        self.build = PageBuild(payload.get("build"))


class ProjectCardEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#projectcardevent
    """

    project_card: ProjectCard
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.project_card = ProjectCard(payload.get("project_card"))
        self.changes = _optional(payload, "changes", RawDict)


class ProjectColumnEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#projectcolumnevent
    """

    project_column: ProjectColumn
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.project_column = ProjectColumn(payload.get("project_column"))
        self.changes = _optional(payload, "changes", RawDict)


class ProjectEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#projectevent
    """

    project: Project

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.project = Project(payload.get("project"))


class PublicEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#publicevent
    """

    def __init__(self, payload: dict):
        super().__init__(payload)


class PullRequestEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#pullrequestevent
    """

    number: int
    pull_request: PullRequest
    assignee: Optional[User]
    label: Optional[Label]
    changes: Optional[RawDict]
    before: Optional[str]
    after: Optional[str]
    requested_reviewer: Optional[User]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.number = payload.get("number")
        self.pull_request = PullRequest(payload.get("pull_request"))
        self.assignee = _optional(payload, "assignee", User)
        self.label = _optional(payload, "label", Label)
        self.changes = _optional(payload, "changes", RawDict)
        self.before = payload.get("before")
        self.after = payload.get("after")
        self.requested_reviewer = _optional(payload, "requested_reviewer", User)


class PullRequestReviewEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#pullrequestreviewevent
    """

    review: Review
    pull_request: PullRequest
    changes: RawDict

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.review = Review(payload.get("review"))
        self.pull_request = PullRequest(payload.get("pull_request"))
        self.changes = RawDict(payload.get("pull_request"))


class PullRequestReviewCommentEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#pullrequestreviewcommentevent
    """

    comment: Comment
    pull_request: PullRequest
    changes: Optional[RawDict]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.comment = Comment(payload.get("comment"))
        self.pull_request = PullRequest(payload.get("pull_request"))
        self.changes = _optional(payload, "changes", RawDict)


class PullRequestReviewThreadEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#pullrequestreviewthreadevent
    """

    pull_request: PullRequest
    thread: Thread

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.pull_request = PullRequest(payload.get("pull_request"))
        self.thread = Thread(payload.get("thread"))


class PushEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#pushevent
    """

    ref: str
    before: str
    after: str
    created: bool
    deleted: bool
    forced: bool
    base_ref: Optional[str]
    compare: str
    commits: List[Commit]
    head_commit: Optional[Commit]
    pusher: CommitUser

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.ref = payload.get("ref")
        self.before = payload.get("before")
        self.after = payload.get("after")
        self.created = payload.get("created")
        self.deleted = payload.get("deleted")
        self.forced = payload.get("forced")
        self.base_ref = payload.get("base_ref")
        self.compare = payload.get("compare")
        self.commits = [Commit(commit) for commit in payload.get("commits")]
        self.head_commit = _optional(payload, "head_commit", Commit)
        self.pusher = CommitUser(payload.get("pusher"))


class ReleaseEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#releaseevent
    """

    release: Release
    changes: RawDict

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.release = Release(payload.get("release"))
        self.changes = RawDict(payload.get("release"))


class RepositoryDispatchEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#repositorydispatchevent
    """

    branch: str
    client_payload: RawDict
    installation: ShortInstallation

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.branch = payload.get("branch")
        self.client_payload = RawDict(payload.get("client_payload"))
        self.installation = ShortInstallation(payload.get("installation"))


class RepositoryEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#repositoryevent
    """

    def __init__(self, payload: dict):
        super().__init__(payload)


class RepositoryImportEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#repositoryimportevent
    """

    status: str

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.status = payload.get("status")


class RepositoryVulnerabilityAlertEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#repositoryvulnerabilityalertevent
    """

    alert: VulnerabilityAlert

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.alert = VulnerabilityAlert(payload.get("alert"))


class SecurityAdvisoryEvent:
    """
    https://developer.github.com/v3/activity/events/types/#securityadvisoryevent
    """

    payload: dict
    action: str
    security_advisory: SecurityAdvisory

    def __init__(self, payload: dict):
        self.payload = payload
        self.action = payload.get("action")
        self.security_advisory = SecurityAdvisory(payload.get("security_advisory"))


class SponsorshipEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#sponsorshipevent
    """

    sponsorship: Sponsorship
    changes: Optional[RawDict]
    effective_date: Optional[str]

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.sponsorship = Sponsorship(payload.get("sponsorship"))
        self.changes = _optional(payload, "changes", RawDict)
        self.effective_date = payload.get("effective_date", None)


class StarEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#starevent
    """

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.starred_at = payload.get("starred_at")


class StatusEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#statusevent
    """

    id: int
    sha: str
    name: str
    target_url: Optional[str]
    avatar_url: Optional[str]
    context: str
    description: Optional[str]
    state: str
    commit: StatusCommit
    branches: List[Branch]
    created_at: str
    updated_at: str

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.id = payload.get("id")
        self.sha = payload.get("sha")
        self.name = payload.get("name")
        self.target_url = payload.get("target_url")
        self.avatar_url = payload.get("avatar_url")
        self.context = payload.get("context")
        self.description = payload.get("description")
        self.state = payload.get("state")
        self.commit = StatusCommit(payload.get("commit"))
        self.branches = [Branch(branch) for branch in payload.get("branches")]
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")


class TeamEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#teamevent
    """

    team: Team

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.team = Team(payload.get("team"))


class TeamAddEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#teamaddevent
    """

    team: Team

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.team = Team(payload.get("team"))


class WatchEvent(BaseWebhookEvent):
    """
    https://developer.github.com/v3/activity/events/types/#watchevent
    """

    def __init__(self, payload: dict):
        super().__init__(payload)


class PingEvent(BaseWebhookEvent):
    """
    https://developer.github.com/webhooks/#ping-event
    """

    zen: str
    hook_id: int
    hook: Hook

    def __init__(self, payload: dict):
        super().__init__(payload)
        self.zen = payload.get("zen")
        self.hook_id = payload.get("hook_id")
        self.hook = Hook(payload.get("hook"))


class WebhookEvent(Enum):
    BRANCH_PROTECTION_RULE = "branch_protection_rule"
    CHECK_RUN = "check_run"
    CHECK_SUITE = "check_suite"
    COMMIT_COMMENT = "commit_comment"
    CONTENT_REFERENCE = "content_reference"
    CREATE = "create"
    DELETE = "delete"
    DEPLOY_KEY = "deploy_key"
    DEPLOYMENT = "deployment"
    DEPLOYMENT_STATUS = "deployment_status"
    FORK = "fork"
    GITHUB_APP_AUTHORIZATION = "github_app_authorization"
    GOLLUM = "gollum"
    INSTALLATION = "installation"
    INSTALLATION_REPOSITORIES = "installation_repositories"
    ISSUE_COMMENT = "issue_comment"
    ISSUES = "issues"
    LABEL = "label"
    MARKETPLACE_PURCHASE = "marketplace_purchase"
    MEMBER = "member"
    MEMBERSHIP = "membership"
    META = "meta"
    MILESTONE = "milestone"
    ORG_BLOCK = "org_block"
    ORGANIZATION = "organization"
    PACKAGE = "package"
    PAGE_BUILD = "page_build"
    PING = "ping"
    PROJECT = "project"
    PUBLIC = "public"
    PULL_REQUEST = "pull_request"
    PULL_REQUEST_REVIEW = "pull_request_review"
    PULL_REQUEST_REVIEW_COMMENT = "pull_request_review_comment"
    PULL_REQUEST_REVIEW_THREAD = "pull_request_review_thread"
    PUSH = "push"
    PROJECT_CARD = "project_card"
    PROJECT_COLUMN = "project_column"
    RELEASE = "release"
    REPOSITORY = "repository"
    REPOSITORY_DISPATCH = "repository_dispatch"
    REPOSITORY_IMPORT = "repository_import"
    REPOSITORY_VULNERABILITY_ALERT = "repository_vulnerability_alert"
    SECURITY_ADVISORY = "security_advisory"
    SPONSORSHIP = "sponsorship"
    STAR = "star"
    STATUS = "status"
    TEAM = "team"
    TEAM_ADD = "team_add"
    WATCH = "watch"


class WebhookEventAction(Enum):
    ADDED = "added"
    ADDED_TO_REPOSITORY = "added_to_repository"
    ARCHIVED = "archived"
    ASSIGNED = "assigned"
    BLOCKED = "blocked"
    CANCELLED = "cancelled"
    CHANGED = "changed"
    CLOSED = "closed"
    COMPLETED = "completed"
    CONVERTED = "converted"
    CONVERTED_TO_DRAFT = "converted_to_draft"
    CREATE = "create"
    CREATED = "created"
    DELETED = "deleted"
    DEMILESTONED = "demilestoned"
    DISMISS = "dismiss"
    DISMISSED = "dismissed"
    EDITED = "edited"
    LABELED = "labeled"
    LOCKED = "locked"
    MEMBER_ADDED = "member_added"
    MEMBER_INVITED = "member_invited"
    MEMBER_REMOVED = "member_removed"
    MILESTONED = "milestoned"
    MOVED = "moved"
    NEW_PERMISSIONS_ACCEPTED = "new_permissions_accepted"
    ON_DEMAND_TEST = "on-demand-test"
    OPENED = "opened"
    PENDING_CANCELLATION = "pending_cancellation"
    PENDING_CHANGE = "pending_change"
    PENDING_CHANGE_CANCELLED = "pending_change_cancelled"
    PENDING_TIER_CHANGE = "pending_tier_change"
    PERFORMED = "performed"
    PINNED = "pinned"
    PRERELEASED = "prereleased"
    PRIVATIZED = "privatized"
    PUBLICIZED = "publicized"
    PUBLISHED = "published"
    PURCHASED = "purchased"
    READY_FOR_REVIEW = "ready_for_review"
    RELEASED = "released"
    REMOVED = "removed"
    REMOVED_FROM_REPOSITORY = "removed_from_repository"
    RENAMED = "renamed"
    REOPENED = "reopened"
    REREQUESTED = "rerequested"
    RESOLVE = "resolve"
    RESOLVED = "resolved"
    REQUESTED = "requested"
    REQUESTED_ACTION = "requested_action"
    REVIEW_REQUESTED = "review_requested"
    REVIEW_REQUEST_REMOVED = "review_request_removed"
    REVOKED = "revoked"
    STARTED = "started"
    SUBMITTED = "submitted"
    SYNCHRONIZE = "synchronize"
    TIER_CHANGED = "tier_changed"
    TRANSFERRED = "transferred"
    UNARCHIVED = "unarchived"
    UNASSIGNED = "unassigned"
    UNBLOCKED = "unblocked"
    UNLABELED = "unlabeled"
    UNLOCKED = "unlocked"
    UNPINNED = "unpinned"
    UNPUBLISHED = "unpublished"
    UNRESOLVED = "unresolved"
    UPDATED = "updated"


event_map = {
    WebhookEvent.BRANCH_PROTECTION_RULE: BranchProtectionRuleEvent,
    WebhookEvent.CHECK_RUN: CheckRunEvent,
    WebhookEvent.CHECK_SUITE: CheckSuiteEvent,
    WebhookEvent.COMMIT_COMMENT: CommitCommentEvent,
    WebhookEvent.CONTENT_REFERENCE: ContentReferenceEvent,
    WebhookEvent.CREATE: CreateEvent,
    WebhookEvent.DELETE: DeleteEvent,
    WebhookEvent.DEPLOY_KEY: DeployKeyEvent,
    WebhookEvent.DEPLOYMENT: DeploymentEvent,
    WebhookEvent.DEPLOYMENT_STATUS: DeploymentStatusEvent,
    WebhookEvent.FORK: ForkEvent,
    WebhookEvent.GITHUB_APP_AUTHORIZATION: GitHubAppAuthorizationEvent,
    WebhookEvent.GOLLUM: GollumEvent,
    WebhookEvent.INSTALLATION: InstallationEvent,
    WebhookEvent.INSTALLATION_REPOSITORIES: InstallationRepositoriesEvent,
    WebhookEvent.ISSUE_COMMENT: IssueCommentEvent,
    WebhookEvent.ISSUES: IssuesEvent,
    WebhookEvent.LABEL: LabelEvent,
    WebhookEvent.MARKETPLACE_PURCHASE: MarketplacePurchaseEvent,
    WebhookEvent.MEMBER: MemberEvent,
    WebhookEvent.MEMBERSHIP: MembershipEvent,
    WebhookEvent.META: MetaEvent,
    WebhookEvent.MILESTONE: MilestoneEvent,
    WebhookEvent.ORG_BLOCK: OrgBlockEvent,
    WebhookEvent.ORGANIZATION: OrganizationEvent,
    WebhookEvent.PACKAGE: PackageEvent,
    WebhookEvent.PAGE_BUILD: PageBuildEvent,
    WebhookEvent.PING: PingEvent,
    WebhookEvent.PROJECT: ProjectEvent,
    WebhookEvent.PUBLIC: PublicEvent,
    WebhookEvent.PULL_REQUEST: PullRequestEvent,
    WebhookEvent.PULL_REQUEST_REVIEW: PullRequestReviewEvent,
    WebhookEvent.PULL_REQUEST_REVIEW_COMMENT: PullRequestReviewCommentEvent,
    WebhookEvent.PULL_REQUEST_REVIEW_THREAD: PullRequestReviewThreadEvent,
    WebhookEvent.PUSH: PushEvent,
    WebhookEvent.PROJECT_CARD: ProjectCardEvent,
    WebhookEvent.PROJECT_COLUMN: ProjectColumnEvent,
    WebhookEvent.RELEASE: ReleaseEvent,
    WebhookEvent.REPOSITORY: RepositoryEvent,
    WebhookEvent.REPOSITORY_DISPATCH: RepositoryDispatchEvent,
    WebhookEvent.REPOSITORY_IMPORT: RepositoryImportEvent,
    WebhookEvent.REPOSITORY_VULNERABILITY_ALERT: RepositoryVulnerabilityAlertEvent,
    WebhookEvent.SECURITY_ADVISORY: SecurityAdvisoryEvent,
    WebhookEvent.SPONSORSHIP: SponsorshipEvent,
    WebhookEvent.STAR: StarEvent,
    WebhookEvent.STATUS: StatusEvent,
    WebhookEvent.TEAM: TeamEvent,
    WebhookEvent.TEAM_ADD: TeamAddEvent,
    WebhookEvent.WATCH: WatchEvent,
}


def parse(event_name, payload: dict):
    return event_map[WebhookEvent(event_name)](payload)



from abc import ABC
from typing import TypeVar, Optional, Type, List, Any

T = TypeVar("T")


def _optional(payload: dict, key: str, class_type: Type[T]) -> Optional[T]:
    if payload.get(key):
        return class_type(payload[key])
    else:
        return None


def _transform(url: str, local_variables: dict) -> str:
    local_variables.pop("self", None)

    for key, value in local_variables.items():
        if not value:
            url = url.split(f"{{/{key}}}")[0]
            # If we find a None value, we shouldn't process the url any more.
            break
        elif f"{{{key}}}" in url:
            url = url.replace(f"{{{key}}}", value)
        elif f"{{/{key}}}" in url:
            url = url.replace(f"{{/{key}}}", f"/{value}")

    return url


class BaseGithubModel(ABC):
    def __new__(cls, *args, **kwargs):
        from octohook import model_overrides

        cls = model_overrides.get(cls) or cls
        return object.__new__(cls)


class Enterprise(BaseGithubModel):
    payload: dict
    id: int
    slug: str
    name: str
    node_id: str
    avatar_url: str
    description: Optional[str]
    website_url: Optional[str]
    html_url: str
    created_at: str
    updated_at: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.slug = payload.get("slug")
        self.name = payload.get("name")
        self.node_id = payload.get("node_id")
        self.avatar_url = payload.get("avatar_url")
        self.description = payload.get("description")
        self.website_url = payload.get("website_url")
        self.html_url = payload.get("html_url")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")

    def __str__(self):
        return self.name


class User(BaseGithubModel):
    payload: dict
    name: Optional[str]
    login: str
    email: Optional[str]
    id: int
    node_id: Optional[str]
    avatar_url: str
    gravatar_id: str
    url: str
    html_url: str
    followers_url: str
    subscriptions_url: str
    organization_url: str
    repos_url: str
    organizations_url: str
    received_events_url: str
    type: str
    site_admin: bool

    def __init__(self, payload: dict):
        self.payload = payload
        self.name = payload.get("name")
        self.login = payload.get("login")
        self.email = payload.get("email")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.avatar_url = payload.get("avatar_url")
        self.gravatar_id = payload.get("gravatar_id")
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.followers_url = payload.get("followers_url")
        self.subscriptions_url = payload.get("subscriptions_url")
        self.organization_url = payload.get("organizations_url")
        self.repos_url = payload.get("repos_url")
        self.organizations_url = payload.get("organizations_url")
        self.received_events_url = payload.get("received_events_url")
        self.type = payload.get("type")
        self.site_admin = payload.get("site_admin")

    def following_url(self, other_user: str = None) -> str:
        return _transform(self.payload["following_url"], locals())

    def gists_url(self, gist_id: str = None) -> str:
        return _transform(self.payload["gists_url"], locals())

    def starred_url(self, owner: str = None, repo: str = None) -> str:
        return _transform(self.payload["starred_url"], locals())

    def events_url(self, privacy: str = None) -> str:
        return _transform(self.payload["events_url"], locals())

    def __str__(self):
        return self.login


class ShortRepository(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    name: str
    full_name: str
    private: bool

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.name = payload.get("name")
        self.full_name = payload.get("full_name")
        self.private = payload.get("private")

    def __str__(self):
        return self.full_name


class RawDict(dict):
    def __init__(self, payload: dict):
        super().__init__(payload)


class Permissions(BaseGithubModel):
    payload: dict
    metadata: str
    contents: str
    issues: str
    administration: str
    statuses: str
    repository_projects: str
    members: str
    repository_hooks: str
    pull_requests: str
    pull: Optional[str]
    push: Optional[str]
    admin: Optional[str]
    pages: str
    deployments: str
    checks: str
    vulnerability_alerts: str
    organization_administration: str
    organization_hooks: str
    organization_plan: str
    organization_projects: str
    organization_user_blocking: str
    team_discussions: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.metadata = payload.get("metadata")
        self.contents = payload.get("contents")
        self.issues = payload.get("issues")
        self.administration = payload.get("administration")
        self.statuses = payload.get("statuses")
        self.repository_projects = payload.get("repository_projects")
        self.members = payload.get("members")
        self.repository_hooks = payload.get("repository_hooks")
        self.pull_requests = payload.get("pull_requests")
        self.pull = payload.get("pull")
        self.push = payload.get("push")
        self.admin = payload.get("admin")
        self.pages = payload.get("pages")
        self.deployments = payload.get("deployments")
        self.checks = payload.get("checks")
        self.vulnerability_alerts = payload.get("vulnerability_alerts")
        self.organization_administration = payload.get("organization_administration")
        self.organization_hooks = payload.get("organization_hooks")
        self.organization_plan = payload.get("organization_plan")
        self.organization_projects = payload.get("organization_projects")
        self.organization_user_blocking = payload.get("organization_user_blocking")
        self.team_discussions = payload.get("team_discussions")


class Repository(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    name: str
    full_name: str
    private: bool
    owner: User
    html_url: str
    description: str
    fork: bool
    url: str
    forks_url: str
    teams_url: str
    hooks_url: str
    events_url: str
    tags_url: str
    languages_url: str
    stargazers_url: str
    contributors_url: str
    subscribers_url: str
    subscription_url: str
    merges_url: str
    downloads_url: str
    deployments_url: str
    created_at: str
    updated_at: str
    pushed_at: str
    git_url: str
    ssh_url: str
    clone_url: str
    svn_url: str
    homepage: Optional[str]
    size: int
    stargazers_count: int
    watchers_count: int
    language: Optional[str]
    has_issues: bool
    has_projects: bool
    has_downloads: bool
    has_wiki: bool
    has_pages: bool
    forks_count: int
    mirror_url: Optional[str]
    archived: bool
    disabled: bool
    open_issues_count: int
    license: Optional[str]
    forks: int
    open_issues: int
    watchers: int
    default_branch: str
    stargazers: Optional[int]
    public: Optional[bool]
    master_branch: Optional[str]
    permissions: Optional[Permissions]
    allow_forking: Optional[bool]
    allow_squash_merge: Optional[bool]
    allow_merge_commit: Optional[bool]
    allow_rebase_merge: Optional[bool]
    allow_update_branch: Optional[bool]
    allow_auto_merge: Optional[bool]
    delete_branch_on_merge: Optional[bool]
    use_squash_pr_title_as_default: Optional[bool]
    squash_merge_commit_message: Optional[str]
    squash_merge_commit_title: Optional[str]
    merge_commit_message: Optional[str]
    merge_commit_title: Optional[str]
    is_template: Optional[bool]
    topics: List[str]
    visibility: Optional[str]
    web_commit_signoff_required: Optional[bool]

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.name = payload.get("name")
        self.full_name = payload.get("full_name")
        self.private = payload.get("private")
        self.owner = User(payload.get("owner"))
        self.html_url = payload.get("html_url")
        self.description = payload.get("description")
        self.fork = payload.get("fork")
        self.url = payload.get("url")
        self.forks_url = payload.get("forks_url")
        self.teams_url = payload.get("teams_url")
        self.hooks_url = payload.get("hooks_url")
        self.events_url = payload.get("events_url")
        self.tags_url = payload.get("tags_url")
        self.languages_url = payload.get("languages_url")
        self.stargazers_url = payload.get("stargazers_url")
        self.contributors_url = payload.get("contributors_url")
        self.subscribers_url = payload.get("subscribers_url")
        self.subscription_url = payload.get("subscription_url")
        self.merges_url = payload.get("merges_url")
        self.downloads_url = payload.get("downloads_url")
        self.deployments_url = payload.get("deployments_url")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.pushed_at = payload.get("pushed_at")
        self.git_url = payload.get("git_url")
        self.ssh_url = payload.get("ssh_url")
        self.clone_url = payload.get("clone_url")
        self.svn_url = payload.get("svn_url")
        self.homepage = payload.get("homepage")
        self.size = payload.get("size")
        self.stargazers_count = payload.get("stargazers_count")
        self.watchers_count = payload.get("watchers_count")
        self.language = payload.get("language")
        self.has_issues = payload.get("has_issues")
        self.has_projects = payload.get("has_projects")
        self.has_downloads = payload.get("has_downloads")
        self.has_wiki = payload.get("has_wiki")
        self.has_pages = payload.get("has_pages")
        self.forks_count = payload.get("forks_count")
        self.mirror_url = payload.get("mirror_url")
        self.archived = payload.get("archived")
        self.disabled = payload.get("disabled")
        self.open_issues_count = payload.get("open_issues_count")
        self.license = payload.get("license")
        self.forks = payload.get("forks")
        self.open_issues = payload.get("open_issues")
        self.watchers = payload.get("watchers")
        self.default_branch = payload.get("default_branch")
        self.stargazers = payload.get("stargazers")
        self.public = payload.get("public")
        self.master_branch = payload.get("master_branch")
        self.permissions = _optional(payload, "permissions", Permissions)
        self.allow_forking = payload.get("allow_forking")
        self.allow_squash_merge = payload.get("allow_squash_merge")
        self.allow_merge_commit = payload.get("allow_merge_commit")
        self.allow_rebase_merge = payload.get("allow_rebase_merge")
        self.allow_auto_merge = payload.get("allow_auto_merge")
        self.allow_update_branch = payload.get("allow_auto_merge")
        self.use_squash_pr_title_as_default = payload.get(
            "use_squash_pr_title_as_default"
        )
        self.squash_merge_commit_message = payload.get("squash_merge_commit_message")
        self.squash_merge_commit_title = payload.get("squash_merge_commit_title")
        self.merge_commit_message = payload.get("merge_commit_message")
        self.merge_commit_title = payload.get("merge_commit_title")
        self.delete_branch_on_merge = payload.get("delete_branch_on_merge")
        self.is_template = payload.get("is_template")
        self.topics = payload.get("topics") or []
        self.visibility = payload.get("visibility")
        self.web_commit_signoff_required = _optional(
            payload, "web_commit_signoff_required", bool
        )

    def keys_url(self, key_id: str = None) -> str:
        return _transform(self.payload["keys_url"], locals())

    def collaborators_url(self, collaborator: str = None) -> str:
        return _transform(self.payload["collaborators_url"], locals())

    def issue_events_url(self, number: int = None) -> str:
        return _transform(self.payload["issue_events_url"], locals())

    def assignees_url(self, user: str = None) -> str:
        return _transform(self.payload["assignees_url"], locals())

    def branches_url(self, branch: str = None) -> str:
        return _transform(self.payload["branches_url"], locals())

    def blobs_url(self, sha: str = None) -> str:
        return _transform(self.payload["blobs_url"], locals())

    def git_tags_url(self, sha: str = None) -> str:
        return _transform(self.payload["git_tags_url"], locals())

    def git_refs_url(self, sha: str = None) -> str:
        return _transform(self.payload["git_refs_url"], locals())

    def trees_url(self, sha: str = None) -> str:
        return _transform(self.payload["trees_url"], locals())

    def statuses_url(self, sha: str) -> str:
        return _transform(self.payload["statuses_url"], locals())

    def commits_url(self, sha: str = None) -> str:
        return _transform(self.payload["commits_url"], locals())

    def git_commits_url(self, sha=None) -> str:
        return _transform(self.payload["git_commits_url"], locals())

    def comments_url(self, number: int = None) -> str:
        return _transform(self.payload["comments_url"], locals())

    def issue_comment_url(self, number: int = None) -> str:
        return _transform(self.payload["issue_comment_url"], locals())

    def contents_url(self, path: str) -> str:
        return _transform(self.payload["contents_url"].replace("+", ""), locals())

    def compare_url(self, base: str, head: str) -> str:
        return _transform(self.payload["compare_url"], locals())

    def archive_url(self, archive_format: str, ref: str = None) -> str:
        return _transform(self.payload["archive_url"], locals())

    def issues_url(self, number: str = None) -> str:
        return _transform(self.payload["issues_url"], locals())

    def pulls_url(self, number: str = None) -> str:
        return _transform(self.payload["pulls_url"], locals())

    def milestones_url(self, number: str = None) -> str:
        return _transform(self.payload["milestones_url"], locals())

    def notifications_url(self, params: str = None) -> str:
        if params is not None:
            if not params.startswith("?"):
                raise AttributeError("Params must be a query string starting with ?")
        replace_with = params if params else ""
        return self.payload["notifications_url"].replace(
            "{?since,all,participating}", replace_with
        )

    def labels_url(self, name: str = None) -> str:
        return _transform(self.payload["labels_url"], locals())

    def releases_url(self, id: str = None) -> str:
        return _transform(self.payload["releases_url"], locals())

    def __str__(self):
        return self.full_name


class Organization(BaseGithubModel):
    payload: dict
    login: str
    id: int
    node_id: str
    url: str
    repos_url: str
    events_url: str
    hooks_url: str
    issues_url: str
    public_members_url: str
    avatar_url: str
    description: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.login = payload.get("login")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.url = payload.get("url")
        self.repos_url = payload.get("repos_url")
        self.events_url = payload.get("events_url")
        self.hooks_url = payload.get("hooks_url")
        self.issues_url = payload.get("issues_url")
        self.public_members_url = payload.get("public_members_url")
        self.avatar_url = payload.get("avatar_url")
        self.description = payload.get("description")

    def members_url(self, member: str = None) -> str:
        return _transform(self.payload["members_url"], locals())

    def public_members_url(self, member: str = None) -> str:
        return _transform(self.payload["public_members_url"], locals())

    def __str__(self):
        return self.login


class Comment(BaseGithubModel):
    payload: dict
    url: str
    html_url: str
    issue_url: Optional[str]
    id: int
    pull_request_review_id: Optional[int]
    original_position: Optional[int]
    original_commit_id: Optional[str]
    pull_request_url: Optional[str]
    diff_hunk: Optional[str]
    node_id: str
    user: User
    position: Optional[int]
    line: Optional[int]
    path: Optional[str]
    commit_id: Optional[str]
    created_at: str
    updated_at: str
    author_association: str
    body: str
    start_line: Optional[int]
    original_start_line: Optional[int]
    start_side: Optional[str]
    original_line: Optional[int]
    side: Optional[str]
    reactions: Optional[RawDict]

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.issue_url = payload.get("issue_url")
        self.id = payload.get("id")
        self.pull_request_review_id = payload.get("pull_request_review_id")
        self.original_position = payload.get("original_position")
        self.original_commit_id = payload.get("original_commit_id")
        self.pull_request_url = payload.get("pull_request_url")
        self.diff_hunk = payload.get("diff_hunk")
        self.node_id = payload.get("node_id")
        self.user = User(payload.get("user"))
        self.position = payload.get("position")
        self.line = payload.get("line")
        self.path = payload.get("path")
        self.commit_id = payload.get("commit_id")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.author_association = payload.get("author_association")
        self.body = payload.get("body")
        self._links = _optional(payload, "_links", RawDict)
        self.start_line = payload.get("start_line")
        self.original_start_line = payload.get("original_start_line")
        self.start_side = payload.get("start_side")
        self.original_line = payload.get("original_line")
        self.side = payload.get("side")
        self.reactions = _optional(payload, "reactions", RawDict)

    def __str__(self):
        return self.body


class Thread(BaseGithubModel):
    payload: dict
    node_id: str
    comments: List[Comment]

    def __init__(self, payload: dict):
        self.payload = payload
        self.node_id = payload.get("node_id")
        self.comments = [Comment(comment) for comment in payload.get("comments")]


class ChecksApp(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    owner: User
    name: str
    description: str
    external_url: str
    html_url: str
    created_at: str
    updated_at: str
    permissions: Permissions
    events: List[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.owner = User(payload.get("owner"))
        self.name = payload.get("name")
        self.description = payload.get("description")
        self.external_url = payload.get("external_url")
        self.html_url = payload.get("html_url")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.permissions = Permissions(payload.get("permissions"))
        self.events = payload.get("events")


class ChecksPullRequest(BaseGithubModel):
    payload: dict
    url: str
    id: int
    number: int
    head: RawDict
    base: RawDict

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.id = payload.get("id")
        self.number = payload.get("number")
        self.head = RawDict(payload.get("head"))
        self.base = RawDict(payload.get("base"))


class CommitUser(BaseGithubModel):
    payload: dict
    name: str
    email: str
    username: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.name = payload.get("name")
        self.email = payload.get("email")
        self.username = payload.get("username")


class Commit(BaseGithubModel):
    payload: dict
    id: str
    tree_id: str
    distinct: bool
    message: str
    url: str
    timestamp: str
    author: CommitUser
    committer: CommitUser
    added: List[str]
    removed: List[str]
    modified: List[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.tree_id = payload.get("tree_id")
        self.distinct = payload.get("distinct")
        self.message = payload.get("message")
        self.url = payload.get("url")
        self.timestamp = payload.get("timestamp")
        self.author = CommitUser(payload.get("author"))
        self.committer = CommitUser(payload.get("committer"))
        self.added = payload.get("added")
        self.removed = payload.get("removed")
        self.modified = payload.get("modified")


class CheckSuite(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    head_branch: str
    head_sha: str
    status: str
    conclusion: Optional[str]
    url: str
    before: str
    after: str
    pull_requests: List[ChecksPullRequest]
    app: ChecksApp
    created_at: str
    updated_at: str
    latest_check_runs_count: Optional[int]
    check_runs_url: Optional[str]
    head_commit: Optional[Commit]

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.head_branch = payload.get("head_branch")
        self.head_sha = payload.get("head_sha")
        self.status = payload.get("status")
        self.conclusion = payload.get("conclusion")
        self.url = payload.get("url")
        self.before = payload.get("before")
        self.after = payload.get("after")
        self.pull_requests = [
            ChecksPullRequest(pr) for pr in payload.get("pull_requests")
        ]
        self.app = ChecksApp(payload.get("app"))
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.latest_check_runs_count = payload.get("latest_check_runs_count")
        self.check_runs_url = payload.get("check_runs_url")
        self.head_commit = _optional(payload, "head_commit", Commit)


class CheckRunOutput(BaseGithubModel):
    payload: dict
    title: Optional[str]
    summary: Optional[str]
    text: Optional[str]
    annotations_count: int
    annotations_url: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.title = payload.get("title")
        self.summary = payload.get("summary")
        self.text = payload.get("text")
        self.annotations_count = payload.get("annotations_count")
        self.annotations_url = payload.get("annotations_url")


class CheckRun(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    head_sha: str
    external_id: str
    url: str
    html_url: str
    details_url: str
    status: str
    conclusion: Optional[str]
    started_at: str
    completed_at: Optional[str]
    output: CheckRunOutput
    name: str
    check_suite: CheckSuite
    app: ChecksApp
    pull_requests: List[ChecksPullRequest]

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.head_sha = payload.get("head_sha")
        self.external_id = payload.get("external_id")
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.details_url = payload.get("details_url")
        self.status = payload.get("status")
        self.conclusion = payload.get("conclusion")
        self.started_at = payload.get("started_at")
        self.completed_at = payload.get("completed_at")
        self.output = CheckRunOutput(payload.get("output"))
        self.name = payload.get("name")
        self.check_suite = CheckSuite(payload.get("check_suite"))
        self.app = ChecksApp(payload.get("app"))
        self.pull_requests = [
            ChecksPullRequest(pr) for pr in payload.get("pull_requests")
        ]


class ShortInstallation(BaseGithubModel):
    payload: dict
    id: int
    node_id: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")


class Installation(BaseGithubModel):
    payload: dict
    id: int
    node_id: Optional[str]
    account: User
    repository_selection: str
    access_tokens_url: str
    repositories_url: str
    html_url: str
    app_id: int
    target_id: int
    permissions: Optional[Permissions]
    events: List[str]
    created_at: int
    updated_at: int
    single_file_name: Optional[str]
    target_type: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id", None)
        self.account = _optional(payload, "account", User)
        self.repository_selection = payload.get("repository_selection")
        self.access_tokens_url = payload.get("access_tokens_url")
        self.repositories_url = payload.get("repositories_url")
        self.html_url = payload.get("html_url")
        self.app_id = payload.get("app_id")
        self.target_id = payload.get("target_id")
        self.permissions = _optional(payload, "permissions", Permissions)
        self.events = payload.get("events")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.single_file_name = payload.get("single_file_name")
        self.target_type = payload.get("target_type")


class DeployKey(BaseGithubModel):
    payload: dict
    id: int
    key: str
    url: str
    title: str
    verified: bool
    created_at: str
    read_only: bool

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.key = payload.get("key")
        self.url = payload.get("url")
        self.title = payload.get("title")
        self.verified = payload.get("verified")
        self.created_at = payload.get("created_at")
        self.read_only = payload.get("read_only")


class Deployment(BaseGithubModel):
    payload: dict
    url: str
    id: int
    node_id: str
    sha: str
    ref: str
    task: str
    payload: RawDict
    original_environment: str
    environment: str
    description: Optional[str]
    creator: User
    created_at: str
    updated_at: str
    statuses_url: str
    repository_url: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.sha = payload.get("sha")
        self.ref = payload.get("ref")
        self.task = payload.get("task")
        self.payload = RawDict(payload.get("payload"))
        self.original_environment = payload.get("original_environment")
        self.environment = payload.get("environment")
        self.description = payload.get("description")
        self.creator = User(payload.get("creator"))
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.statuses_url = payload.get("statuses_url")
        self.repository_url = payload.get("repository_url")


class DeploymentStatus(BaseGithubModel):
    payload: dict
    url: str
    id: int
    node_id: str
    state: str
    creator: User
    description: str
    environment: str
    target_url: str
    created_at: str
    updated_at: str
    deployment_url: str
    repository_url: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.state = payload.get("state")
        self.creator = User(payload.get("creator"))
        self.description = payload.get("description")
        self.environment = payload.get("environment")
        self.target_url = payload.get("target_url")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.deployment_url = payload.get("deployment_url")
        self.repository_url = payload.get("repository_url")


class Page(BaseGithubModel):
    payload: dict
    page_name: str
    title: str
    summary: str
    action: str
    sha: str
    html_url: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.page_name = payload.get("page_name")
        self.title = payload.get("title")
        self.summary = payload.get("summary")
        self.action = payload.get("action")
        self.sha = payload.get("sha")
        self.html_url = payload.get("html_url")


class Label(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    url: str
    name: str
    color: str
    description: str
    default: bool

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.url = payload.get("url")
        self.name = payload.get("name")
        self.color = payload.get("color")
        self.description = payload.get("description")
        self.default = payload.get("default")

    def __eq__(self, other):
        return self.name == other

    def __repr__(self):
        return f"{self.name}(#{self.color})"

    def __str__(self):
        return self.name


class Milestone(BaseGithubModel):
    payload: dict
    url: str
    html_url: str
    labels_url: str
    id: int
    node_id: str
    number: int
    title: str
    description: str
    creator: Optional[User]
    open_issues: int
    closed_issues: int
    state: str
    created_at: str
    updated_at: str
    due_on: Optional[str]
    closed_at: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.labels_url = payload.get("labels_url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.number = payload.get("number")
        self.title = payload.get("title")
        self.description = payload.get("description")
        self.creator = _optional(payload, "creator", User)
        self.open_issues = payload.get("open_issues")
        self.closed_issues = payload.get("closed_issues")
        self.state = payload.get("state")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.due_on = payload.get("due_on")
        self.closed_at = payload.get("closed_at")


class Issue(BaseGithubModel):
    payload: dict
    url: str
    repository_url: str
    comments_url: str
    events_url: str
    html_url: str
    id: int
    node_id: str
    number: int
    title: str
    user: User
    labels: List[Label]
    state: str
    locked: bool
    assignee: Optional[User]
    assignees: List[User]
    milestone: Optional[Milestone]
    comments: int
    created_at: str
    updated_at: str
    closed_at: Optional[str]
    author_association: str
    body: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.repository_url = payload.get("repository_url")
        self.comments_url = payload.get("comments_url")
        self.events_url = payload.get("events_url")
        self.html_url = payload.get("html_url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.number = payload.get("number")
        self.title = payload.get("title")
        self.user = User(payload.get("user"))
        self.labels = [Label(label) for label in payload.get("labels")]
        self.state = payload.get("state")
        self.locked = payload.get("locked")
        self.assignee = _optional(payload, "assignee", User)
        self.assignees = [User(assignee) for assignee in payload.get("assignees")]
        self.milestone = _optional(payload, "milestone", Milestone)
        self.comments = payload.get("comments")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.closed_at = payload.get("closed_at")
        self.author_association = payload.get("author_association")
        self.body = payload.get("body")

    def labels_url(self, name: str = None) -> str:
        return _transform(self.payload["labels_url"], locals())


class PurchaseAccount(BaseGithubModel):
    payload: dict
    type: str
    id: int
    login: str
    organization_billing_email: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.type = payload.get("type")
        self.id = payload.get("id")
        self.login = payload.get("login")
        self.organization_billing_email = payload.get("organization_billing_email")


class Plan(BaseGithubModel):
    payload: dict
    id: int
    name: str
    description: str
    monthly_price_in_cents: int
    yearly_price_in_cents: int
    yearly_price: Optional[int]
    price_model: str
    has_free_trial: bool
    unit_name: str
    bullets: List[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.name = payload.get("name")
        self.description = payload.get("description")
        self.monthly_price_in_cents = payload.get("monthly_price_in_cents")
        self.yearly_price_in_cents = payload.get("yearly_price_in_cents")
        self.yearly_price = payload.get("yearly_price")
        self.price_model = payload.get("price_model")
        self.has_free_trial = payload.get("has_free_trial")
        self.unit_name = payload.get("unit_name")
        self.bullets = payload.get("bullets")


class MarketplacePurchase(BaseGithubModel):
    payload: dict
    account: PurchaseAccount
    billing_cycle: str
    unit_count: int
    on_free_trial: bool
    free_trial_ends_on: Optional[str]
    next_billing_date: str
    plan: Plan

    def __init__(self, payload: dict):
        self.payload = payload
        self.account = PurchaseAccount(payload.get("account"))
        self.billing_cycle = payload.get("billing_cycle")
        self.unit_count = payload.get("unit_count")
        self.on_free_trial = payload.get("on_free_trial")
        self.free_trial_ends_on = payload.get("free_trial_ends_on")
        self.next_billing_date = payload.get("next_billing_date")
        self.plan = Plan(payload.get("plan"))


class Team(BaseGithubModel):
    payload: dict
    name: str
    id: int
    node_id: str
    slug: str
    description: str
    privacy: str
    url: str
    html_url: str
    repositories_url: str
    permission: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.name = payload.get("name")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.slug = payload.get("slug")
        self.description = payload.get("description")
        self.privacy = payload.get("privacy")
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.repositories_url = payload.get("repositories_url")
        self.permission = payload.get("permission")

    def members_url(self, member: str = None) -> str:
        return _transform(self.payload["members_url"], locals())


class Hook(BaseGithubModel):
    payload: dict
    type: str
    id: int
    name: str
    active: bool
    events: List[str]
    config: RawDict
    updated_at: str
    created_at: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.type = payload.get("type")
        self.id = payload.get("id")
        self.name = payload.get("name")
        self.active = payload.get("active")
        self.events = payload.get("events")
        self.config = RawDict(payload.get("config"))
        self.updated_at = payload.get("updated_at")
        self.created_at = payload.get("created_at")


class Membership(BaseGithubModel):
    payload: dict
    url: str
    state: str
    role: str
    organization_url: str
    user: User

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.state = payload.get("state")
        self.role = payload.get("role")
        self.organization_url = payload.get("organization_url")
        self.user = User(payload.get("user"))


class Asset(BaseGithubModel):
    payload: dict
    url: str
    id: str
    node_id: str
    name: str
    label: str
    uploader: User
    content_type: str
    state: str
    size: str
    download_count: str
    created_at: str
    updated_at: str
    browser_download_url: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.name = payload.get("name")
        self.label = payload.get("label")
        self.uploader = User(payload.get("uploader"))
        self.content_type = payload.get("content_type")
        self.state = payload.get("state")
        self.size = payload.get("size")
        self.download_count = payload.get("download_count")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.browser_download_url = payload.get("browser_download_url")


class Release(BaseGithubModel):
    payload: dict
    url: str
    assets_url: Optional[str]
    upload_url: Optional[str]
    html_url: str
    id: int
    node_id: Optional[str]
    tag_name: str
    target_commitish: str
    name: str
    draft: bool
    author: User
    prerelease: bool
    created_at: str
    published_at: Optional[str]
    assets: List[Asset]
    tarball_url: Optional[str]
    zipball_url: Optional[str]
    body: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.assets_url = payload.get("assets_url", None)
        self.upload_url = payload.get("upload_url", None)
        self.html_url = payload.get("html_url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id", None)
        self.tag_name = payload.get("tag_name")
        self.target_commitish = payload.get("target_commitish")
        self.name = payload.get("name")
        self.draft = payload.get("draft")
        self.author = User(payload.get("author"))
        self.prerelease = payload.get("prerelease")
        self.created_at = payload.get("created_at")
        self.published_at = payload.get("published_at")
        self.assets = [Asset(asset) for asset in payload.get("assets", [])]
        self.tarball_url = payload.get("tarball_url", None)
        self.zipball_url = payload.get("zipball_url", None)
        self.body = payload.get("body", None)


class PackageFile(BaseGithubModel):
    payload: dict
    download_url: str
    id: str
    name: str
    sha256: str
    sha1: str
    md5: str
    content_type: str
    state: str
    size: str
    created_at: str
    updated_at: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.download_url = payload.get("download_url")
        self.id = payload.get("id")
        self.name = payload.get("name")
        self.sha256 = payload.get("sha256")
        self.sha1 = payload.get("sha1")
        self.md5 = payload.get("md5")
        self.content_type = payload.get("content_type")
        self.state = payload.get("state")
        self.size = payload.get("size")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")


class PackageVersion(BaseGithubModel):
    payload: dict
    id: int
    version: str
    summary: str
    body: str
    body_html: str
    release: Release
    manifest: str
    html_url: str
    tag_name: str
    target_commitish: str
    target_oid: str
    draft: bool
    prerelease: bool
    created_at: str
    updated_at: str
    metadata: List[Any]
    package_files: List[PackageFile]
    author: User
    installation_command: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.version = payload.get("version")
        self.summary = payload.get("summary")
        self.body = payload.get("body")
        self.body_html = payload.get("body_html")
        self.release = Release(payload.get("release"))
        self.manifest = payload.get("manifest")
        self.html_url = payload.get("html_url")
        self.tag_name = payload.get("tag_name")
        self.target_commitish = payload.get("target_commitish")
        self.target_oid = payload.get("target_oid")
        self.draft = payload.get("draft")
        self.prerelease = payload.get("prerelease")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.metadata = payload.get("metadata")
        self.package_files = [
            PackageFile(file) for file in payload.get("package_files")
        ]
        self.author = User(payload.get("author"))
        self.installation_command = payload.get("installation_command")


class Registry(BaseGithubModel):
    payload: dict
    about_url: str
    name: str
    type: str
    url: str
    vendor: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.about_url = payload.get("about_url")
        self.name = payload.get("name")
        self.type = payload.get("type")
        self.url = payload.get("url")
        self.vendor = payload.get("vendor")


class Package(BaseGithubModel):
    payload: dict
    id: int
    name: str
    package_type: str
    html_url: str
    created_at: str
    updated_at: str
    owner: User
    package_version: PackageVersion
    registry: Registry

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.name = payload.get("name")
        self.package_type = payload.get("package_type")
        self.html_url = payload.get("html_url")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.owner = User(payload.get("owner"))
        self.package_version = PackageVersion(payload.get("package_version"))
        self.registry = Registry(payload.get("registry"))


class PageBuild(BaseGithubModel):
    payload: dict
    url: str
    status: str
    error: RawDict
    pusher: User
    commit: str
    duration: int
    created_at: str
    updated_at: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.status = payload.get("status")
        self.error = RawDict(payload.get("error"))
        self.pusher = User(payload.get("pusher"))
        self.commit = payload.get("commit")
        self.duration = payload.get("duration")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")


class ProjectCard(BaseGithubModel):
    payload: dict
    url: str
    project_url: str
    column_url: str
    column_id: int
    id: int
    node_id: str
    note: Optional[str]
    archived: bool
    after_id: Optional[str]
    creator: User
    created_at: str
    updated_at: str
    content_url: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.project_url = payload.get("project_url")
        self.column_url = payload.get("column_url")
        self.column_id = payload.get("column_id")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.note = payload.get("note")
        self.archived = payload.get("archived")
        self.after_id = payload.get("after_id")
        self.creator = User(payload.get("creator"))
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.content_url = payload.get("content_url")


class ProjectColumn(BaseGithubModel):
    payload: dict
    url: str
    project_url: str
    cards_url: str
    id: int
    node_id: str
    name: str
    created_at: str
    updated_at: str
    after_id: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.project_url = payload.get("project_url")
        self.cards_url = payload.get("cards_url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.name = payload.get("name")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.after_id = payload.get("after_id")


class Project(BaseGithubModel):
    payload: dict
    owner_url: str
    url: str
    html_url: str
    columns_url: str
    id: int
    node_id: str
    name: str
    body: str
    number: int
    state: str
    creator: User
    created_at: str
    updated_at: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.owner_url = payload.get("owner_url")
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.columns_url = payload.get("columns_url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.name = payload.get("name")
        self.body = payload.get("body")
        self.number = payload.get("number")
        self.state = payload.get("state")
        self.creator = User(payload.get("creator"))
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")


class Ref(BaseGithubModel):
    payload: dict
    label: str
    ref: str
    sha: str
    user: User
    repo: Repository

    def __init__(self, payload: dict):
        self.payload = payload
        self.label = payload.get("label")
        self.ref = payload.get("ref")
        self.sha = payload.get("sha")
        self.user = User(payload.get("user"))
        self.repo = Repository(payload.get("repo"))


class PullRequest(BaseGithubModel):
    payload: dict
    url: str
    id: int
    node_id: str
    html_url: str
    diff_url: str
    patch_url: str
    issue_url: str
    number: int
    state: str
    locked: bool
    title: str
    user: User
    body: str
    created_at: str
    updated_at: str
    closed_at: Optional[str]
    merged_at: Optional[str]
    merge_commit_sha: Optional[str]
    assignee: Optional[User]
    assignees: List[User]
    requested_reviewers: List[User]
    requested_teams: List[str]
    labels: List[Label]
    milestone: Optional[str]
    commits_url: str
    review_comments_url: str
    comments_url: str
    statuses_url: str
    head: Ref
    base: Ref
    _links: RawDict
    author_association: str
    draft: bool
    merged: Optional[bool]
    mergeable: Optional[bool]
    rebaseable: Optional[bool]
    mergeable_state: Optional[str]
    merged_by: Optional[str]
    comments: Optional[int]
    review_comments: Optional[int]
    maintainer_can_modify: Optional[bool]
    commits: Optional[int]
    additions: Optional[int]
    deletions: Optional[int]
    changed_files: Optional[int]
    auto_merge: Optional[bool]
    active_lock_reason: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.html_url = payload.get("html_url")
        self.diff_url = payload.get("diff_url")
        self.patch_url = payload.get("patch_url")
        self.issue_url = payload.get("issue_url")
        self.number = payload.get("number")
        self.state = payload.get("state")
        self.locked = payload.get("locked")
        self.title = payload.get("title")
        self.user = User(payload.get("user"))
        self.body = payload.get("body")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.closed_at = payload.get("closed_at")
        self.merged_at = payload.get("merged_at")
        self.merge_commit_sha = payload.get("merge_commit_sha")
        self.assignee = _optional(payload, "assignee", User)
        self.assignees = [User(assignee) for assignee in payload.get("assignees")]
        self.requested_reviewers = [
            User(reviewer) for reviewer in payload.get("requested_reviewers")
        ]
        self.requested_teams = payload.get("requested_teams")
        self.labels = [Label(item) for item in payload.get("labels")]
        self.milestone = payload.get("milestone")
        self.commits_url = payload.get("commits_url")
        self.review_comments_url = payload.get("review_comments_url")
        self.comments_url = payload.get("comments_url")
        self.statuses_url = payload.get("statuses_url")
        self.head = Ref(payload.get("head"))
        self.base = Ref(payload.get("base"))
        self._links = RawDict(payload.get("_links"))
        self.author_association = payload.get("author_association")
        self.draft = payload.get("draft")
        self.merged = payload.get("merged")
        self.mergeable = payload.get("mergeable")
        self.rebaseable = payload.get("rebaseable")
        self.mergeable_state = payload.get("mergeable_state")
        self.merged_by = payload.get("merged_by")
        self.comments = payload.get("comments")
        self.review_comments = payload.get("review_comments")
        self.maintainer_can_modify = payload.get("maintainer_can_modify")
        self.commits = payload.get("commits")
        self.additions = payload.get("additions")
        self.deletions = payload.get("deletions")
        self.changed_files = payload.get("changed_files")
        self.auto_merge = payload.get("auto_merge")
        self.active_lock_reason = payload.get("active_lock_reason")

    def review_comment_url(self, number: int = None) -> str:
        return _transform(self.payload["review_comment_url"], locals())

    def __str__(self):
        return f"#{self.number} {self.title}"


class Review(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    user: User
    body: Optional[str]
    commit_id: str
    submitted_at: str
    state: str
    html_url: str
    pull_request_url: str
    author_association: str
    _links: RawDict

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.user = User(payload.get("user"))
        self.body = payload.get("body")
        self.commit_id = payload.get("commit_id")
        self.submitted_at = payload.get("submitted_at")
        self.state = payload.get("state")
        self.html_url = payload.get("html_url")
        self.pull_request_url = payload.get("pull_request_url")
        self.author_association = payload.get("author_association")
        self._links = RawDict(payload.get("_links"))


class VulnerabilityAlert(BaseGithubModel):
    payload: dict
    id: int
    affected_range: str
    affected_package_name: str
    external_reference: str
    external_identifier: str
    fixed_in: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.affected_range = payload.get("affected_range")
        self.affected_package_name = payload.get("affected_package_name")
        self.external_reference = payload.get("external_reference")
        self.external_identifier = payload.get("external_identifier")
        self.fixed_in = payload.get("fixed_in")


class VulnerablePackage(BaseGithubModel):
    payload: dict
    ecosystem: str
    name: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.ecosystem = payload.get("ecosystem")
        self.name = payload.get("name")


class PackageVersionInfo(BaseGithubModel):
    payload: dict
    identifier: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.identifier = payload.get("identifier")


class Vulnerability(BaseGithubModel):
    payload: dict
    package: VulnerablePackage
    severity: str
    vulnerable_version_range: str
    first_patched_version: PackageVersionInfo

    def __init__(self, payload: dict):
        self.payload = payload
        self.package = VulnerablePackage(payload.get("package"))
        self.severity = payload.get("severity")
        self.vulnerable_version_range = payload.get("vulnerable_version_range")
        self.first_patched_version = PackageVersionInfo(
            payload.get("first_patched_version")
        )


class SecurityVulnerabilityIdentifier(BaseGithubModel):
    payload: dict
    value: str
    type: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.value: str = payload.get("value")
        self.type: str = payload.get("type")


class SecurityAdvisoryReference(BaseGithubModel):
    payload: dict
    url: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.url = payload.get("url")


class SecurityAdvisory(BaseGithubModel):
    payload: dict
    ghsa_id: str
    summary: str
    description: str
    severity: str
    identifiers: List[SecurityVulnerabilityIdentifier]
    references: List[SecurityAdvisoryReference]
    published_at: str
    updated_at: str
    withdrawn_at: Optional[str]
    vulnerabilities: List[Vulnerability]

    def __init__(self, payload: dict):
        self.payload = payload
        self.ghsa_id = payload.get("ghsa_id")
        self.summary = payload.get("summary")
        self.description = payload.get("description")
        self.severity = payload.get("severity")
        self.identifiers = [
            SecurityVulnerabilityIdentifier(identifier)
            for identifier in payload.get("identifiers")
        ]
        self.references = [
            SecurityAdvisoryReference(item) for item in payload.get("references")
        ]
        self.published_at = payload.get("published_at")
        self.updated_at = payload.get("updated_at")
        self.withdrawn_at = payload.get("withdrawn_at")
        self.vulnerabilities = [
            Vulnerability(vulnerability)
            for vulnerability in payload.get("vulnerabilities")
        ]


class SponsorshipTier(BaseGithubModel):
    payload: dict
    node_id: str
    created_at: str
    description: str
    monthly_price_in_cents: int
    monthly_price_in_dollars: int
    name: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.node_id = payload.get("node_id")
        self.created_at = payload.get("created_at")
        self.description = payload.get("description")
        self.monthly_price_in_cents = payload.get("monthly_price_in_cents")
        self.monthly_price_in_dollars = payload.get("monthly_price_in_dollars")
        self.name = payload.get("name")


class Sponsorship(BaseGithubModel):
    payload: dict
    node_id: str
    created_at: str
    maintainer: User
    sponsor: User
    privacy_level: str
    tier: SponsorshipTier

    def __init__(self, payload: dict):
        self.payload = payload
        self.node_id = payload.get("node_id")
        self.created_at = payload.get("created_at")
        self.maintainer = User(payload.get("maintainer"))
        self.sponsor = User(payload.get("sponsor"))
        self.privacy_level = payload.get("privacy_level")
        self.tier = SponsorshipTier(payload.get("tier"))


class StatusBranchCommit(BaseGithubModel):
    payload: dict
    sha: str
    url: str
    html_url: Optional[str]

    def __init__(self, payload: dict):
        self.payload = payload
        self.sha = payload.get("sha")
        self.url = payload.get("url")
        self.html_url = payload.get("html_url", None)


class Branch(BaseGithubModel):
    payload: dict
    name: str
    commit: StatusBranchCommit
    protected: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.name = payload.get("name")
        self.commit = StatusBranchCommit(payload.get("commit"))
        self.protected = payload.get("protected")


class StatusCommitVerification(BaseGithubModel):
    payload: dict
    verified: bool
    reason: str
    signature: str
    payload: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.verified = payload.get("verified")
        self.reason = payload.get("reason")
        self.signature = payload.get("signature")
        self.payload = payload.get("payload")


class StatusNestedCommitUser(BaseGithubModel):
    payload: dict
    name: str
    email: str
    date: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.name = payload.get("name")
        self.email = payload.get("email")
        self.date = payload.get("date")


class StatusNestedCommit(BaseGithubModel):
    payload: dict
    author: StatusNestedCommitUser
    committer: StatusNestedCommitUser
    message: str
    tree: StatusBranchCommit
    url: str
    comment_count: int
    verification: StatusCommitVerification

    def __init__(self, payload: dict):
        self.payload = payload
        self.author = StatusNestedCommitUser(payload.get("author"))
        self.committer = StatusNestedCommitUser(payload.get("committer"))
        self.message = payload.get("message")
        self.tree = StatusBranchCommit(payload.get("tree"))
        self.url = payload.get("url")
        self.comment_count = payload.get("comment_count")
        self.verification = StatusCommitVerification(payload.get("verification"))


class StatusCommit(BaseGithubModel):
    payload: dict
    sha: str
    node_id: str
    commit: StatusNestedCommit
    url: str
    html_url: str
    comments_url: str
    author: Optional[User]
    committer: Optional[User]
    parents: List[StatusBranchCommit]

    def __init__(self, payload: dict):
        self.payload = payload
        self.sha = payload.get("sha")
        self.node_id = payload.get("node_id")
        self.commit = StatusNestedCommit(payload.get("commit"))
        self.url = payload.get("url")
        self.html_url = payload.get("html_url")
        self.comments_url = payload.get("comments_url")
        self.author = _optional(payload, "author", User)
        self.committer = _optional(payload, "committer", User)
        self.parents = [StatusBranchCommit(parent) for parent in payload.get("parents")]


class ContentReference(BaseGithubModel):
    payload: dict
    id: int
    node_id: str
    reference: str

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.node_id = payload.get("node_id")
        self.reference = payload.get("reference")


class Rule(BaseGithubModel):
    payload: dict
    id: int
    repository_id: int
    name: str
    created_at: str
    updated_at: str
    pull_request_reviews_enforcement_level: str
    required_approving_review_count: int
    dismiss_stale_reviews_on_push: bool
    require_code_owner_review: bool
    authorized_dismissal_actors_only: bool
    ignore_approvals_from_contributors: bool
    required_status_checks: List[str]
    required_status_checks_enforcement_level: str
    strict_required_status_checks_policy: bool
    signature_requirement_enforcement_level: str
    linear_history_requirement_enforcement_level: str
    admin_enforced: bool
    create_protected: bool
    allow_force_pushes_enforcement_level: str
    allow_deletions_enforcement_level: str
    merge_queue_enforcement_level: str
    required_deployments_enforcement_level: str
    required_conversation_resolution_level: str
    authorized_actors_only: bool
    authorized_actor_names: List[str]
    require_last_push_approval: bool

    def __init__(self, payload: dict):
        self.payload = payload
        self.id = payload.get("id")
        self.repository_id = payload.get("repository_id")
        self.name = payload.get("name")
        self.created_at = payload.get("created_at")
        self.updated_at = payload.get("updated_at")
        self.pull_request_reviews_enforcement_level = payload.get(
            "pull_request_reviews_enforcement_level"
        )
        self.required_approving_review_count = payload.get(
            "required_approving_review_count"
        )
        self.dismiss_stale_reviews_on_push = payload.get(
            "dismiss_stale_reviews_on_push"
        )
        self.require_code_owner_review = payload.get("require_code_owner_review")
        self.authorized_dismissal_actors_only = payload.get(
            "authorized_dismissal_actors_only"
        )
        self.ignore_approvals_from_contributors = payload.get(
            "ignore_approvals_from_contributors"
        )
        self.required_status_checks = payload.get("required_status_checks")
        self.required_status_checks_enforcement_level = payload.get(
            "required_status_checks_enforcement_level"
        )
        self.strict_required_status_checks_policy = payload.get(
            "strict_required_status_checks_policy"
        )
        self.signature_requirement_enforcement_level = payload.get(
            "signature_requirement_enforcement_level"
        )
        self.linear_history_requirement_enforcement_level = payload.get(
            "linear_history_requirement_enforcement_level"
        )
        self.admin_enforced = payload.get("admin_enforced")
        self.create_protected = payload.get("create_protected")
        self.allow_force_pushes_enforcement_level = payload.get(
            "allow_force_pushes_enforcement_level"
        )
        self.allow_deletions_enforcement_level = payload.get(
            "allow_deletions_enforcement_level"
        )
        self.merge_queue_enforcement_level = payload.get(
            "merge_queue_enforcement_level"
        )
        self.required_deployments_enforcement_level = payload.get(
            "required_deployments_enforcement_level"
        )
        self.required_conversation_resolution_level = payload.get(
            "required_conversation_resolution_level"
        )
        self.authorized_actors_only = payload.get("authorized_actors_only")
        self.authorized_actor_names = payload.get("authorized_actor_names")
        self.require_last_push_approval = payload.get("require_last_push_approval")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="octohook",
    version="0.12.0",
    author="Sridhar Vadlamani",
    author_email="v.sridhar.sreenivas@gmail.com",
    description="Typed interactions with Github Webhooks",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/doodla/octohook",
    packages=["octohook"],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Development Status :: 4 - Beta",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.7",
)



import logging
from collections import defaultdict
from functools import wraps
from typing import List

from octohook.events import WebhookEvent, WebhookEventAction, parse

logger = logging.getLogger("octohook")

DEBUG = "debug"
ANY_REPO = "*"
ANY_ACTION = "*"


class _WebhookDecorator:
    def __init__(self):
        self.handlers = defaultdict(  # event_name
            lambda: defaultdict(lambda: defaultdict(set))  # action  # repo  # handlers
        )

    def webhook(
        self,
        event: WebhookEvent,
        actions: List[WebhookEventAction] = None,
        repositories: List[str] = None,
        debug=False,
    ):
        def real_decorator(fn):
            @wraps(fn)
            def wrapper(*, event_name, payload):
                fn(parse(event_name, payload))

            repos = repositories or []

            if debug:
                if not repos:
                    self.handlers[event][DEBUG][ANY_REPO].add(wrapper)
                else:
                    for repo in repos:
                        self.handlers[event][DEBUG][repo].add(wrapper)

            if not actions:
                if not repos:
                    self.handlers[event][ANY_ACTION][ANY_REPO].add(wrapper)
                else:
                    for repo in repos:
                        self.handlers[event][ANY_ACTION][repo].add(wrapper)
            else:
                for action in actions:
                    if not repos:
                        self.handlers[event][action][ANY_REPO].add(wrapper)
                    else:
                        for repo in repos:
                            self.handlers[event][action][repo].add(wrapper)

            return wrapper

        return real_decorator

    def handle_webhook(self, event_name: str, payload: dict):
        action_handlers = self.handlers[WebhookEvent(event_name)]

        # These handlers are invoked all the time.
        handlers = action_handlers[ANY_ACTION][ANY_REPO].copy()

        repo_name = payload["repository"]["full_name"]

        if payload.get("action") is not None:
            handlers.update(
                action_handlers[WebhookEventAction(payload["action"])][ANY_REPO]
            )
            handlers.update(
                action_handlers[WebhookEventAction(payload["action"])][repo_name]
            )

        if action_handlers[DEBUG]:
            logger.info("Debug handlers found.")
            handlers = action_handlers[DEBUG][ANY_REPO].copy()
            handlers.update(action_handlers[DEBUG][repo_name])

        for handler in handlers:
            # noinspection PyBroadException
            try:
                logger.info(f"Evaluating {handler.__name__}")
                handler(event_name=event_name, payload=payload)
            except Exception:
                logger.exception(f"Exception when handling {handler.__name__}")


_decorator = _WebhookDecorator()

hook = _decorator.webhook
handle_webhook = _decorator.handle_webhook



from importlib import import_module
from pathlib import Path
from pkgutil import walk_packages
from typing import List

import octohook.models
from .decorators import hook, handle_webhook
from .events import parse, WebhookEvent, WebhookEventAction

__all__ = [
    "events",
    "handle_webhook",
    "hook",
    "load_hooks",
    "models",
    "model_overrides",
    "parse",
    "WebhookEvent",
    "WebhookEventAction",
]

_imported_modules = []
model_overrides = {}


def _import_module(module: str) -> List[str]:
    module_path = import_module(module).__file__

    package_dir = Path(module_path).resolve().parent

    imported_modules = []
    for _, module_name, is_package in walk_packages([str(package_dir)]):
        module_to_be_imported = f"{module}.{module_name}"
        if is_package:
            imported_modules.extend(_import_module(module_to_be_imported))
        else:
            imported_modules.append(import_module(module_to_be_imported).__name__)
    return imported_modules


def load_hooks(modules: List[str]):
    """
    Iterates through the given modules recursively and imports all the modules to load hook information.

    @param modules List of modules to be imported. The modules cannot be relative. For example, you may use
                   "module_a.module_b" or "module_a", but not ".module_a" or "module_a/module_b"
    """
    global _imported_modules

    for module in modules:
        _imported_modules.extend(_import_module(module))
